#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <dirent.h>
#include <unistd.h>

#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/reboot.h>

#include <asm/types.h>
#include <linux/netlink.h>

#define UEVENT_BUFFER_SIZE	2048

static int mount_sys_fs(const char *fsname, const char *fstype)
{
	mkdir(fsname, 0555);

	if (mount("none", fsname, fstype, 0, NULL) == -1)
	{
		return -1;
	}

	return 1;
}

int netlink_socket(void)
{
	char    buf[UEVENT_BUFFER_SIZE];
	ssize_t buflen;
	ssize_t index;
	fd_set  readfds;
	struct  sockaddr_nl snl;
	int     sock;

	memset(&snl, 0, sizeof snl);
	snl.nl_family = AF_NETLINK;
	snl.nl_pid    = getpid();
	snl.nl_groups = 1;

	sock = socket(AF_NETLINK, SOCK_DGRAM, NETLINK_KOBJECT_UEVENT);
	if (sock == -1) {
		perror("socket:");
		return 1;
	}

	if (bind(sock, (struct sockaddr *) &snl,(ssize_t) sizeof snl) == -1) {
		perror("bind:");
		close(sock);
		sock = -1;
		return 1;
	}

	while (1) {
		FD_ZERO(&readfds);
		FD_SET(sock, &readfds);

		int handle_count = select(sock + 1, &readfds, NULL, NULL, NULL);
		if (handle_count == -1) {
			return 1;
		}

		if (FD_ISSET(sock, &readfds)) {
			buflen = recv(sock, &buf, sizeof(buf), 0);
			if (buflen <= 0) {
				perror("recv:");
				return 0;
			}

			if (strcmp(buf, CMDEJECT) == 0)
				reboot(RB_AUTOBOOT);
		}
	}
}

int main(int argc, char **argv)
{
	int    mnt_procfs;
	int    mnt_sysfs;
	FILE   *fp;
	char   blkdev[]    = "/dev/mmcblk0\x0";
	char   blktmp[200] = "\x0";

	printf("init built on %s at %s\n", __DATE__, __TIME__);

	/* Mount /sys */
	mnt_sysfs = mount_sys_fs("/sys", "sysfs") >= 0;
	if (!mnt_sysfs)
	{
		printf("could not mount sysfs\n");
			exit(EXIT_FAILURE);
	}

	/* Mount /proc */
	mnt_procfs = mount_sys_fs("/proc", "proc") >= 0;
	if (!mnt_procfs)
	{
		printf("could not mount %s as %s\n", "/proc/cmdline", "/proc");
		return 1;
	}

	sleep(2);	// 1 Works but it's safer this way.

/*	DIR *dir=opendir("/sys/devices/lm-2/gadget/");
	struct dirent * dp;

	while ((dp = readdir (dir))!= NULL) {
		  printf ("%s \n", dp->d_name);
	  }
	closedir(dir);*/
	
	fp = fopen(GADGETPATH, "wb");
	if (fp == NULL)
	{
		printf ("Unable to open gadget-lun0 on path %s (wb).\n",GADGETPATH);
		return 1;
	}

	if (fwrite (blkdev, 1, strlen(blkdev), fp) != strlen(blkdev))
	{
		printf ("Unable to write bind block device to usb file storage\n");
		return 1;
	}

	fclose (fp);

	fp = fopen(GADGETPATH, "rb");
	if (fp == NULL)
	{
		printf ("Unable to open gadget-lun0 on path %s (rb).\n",GADGETPATH);
		return 1;
	}

	fread(blktmp, 200, 1, fp);
	fclose (fp);
	
	if (strncmp (blktmp, blkdev, strlen(blkdev)) != 0)
	{
		printf ("Unable to bind block device %s to mass storage\n", blkdev);
		return 1;
	}

#ifdef WATCHDOGPATH
	/* Write to the watchdog. Since this is an irvine thing, do not bitch if it doesn't exist */
	fp = fopen(WATCHDOGPATH, "wb");
	if (fp != NULL) {
		printf("Beanbag!...\n");
		fwrite("V\n", strlen("V\n")+1, 1, fp);
		fflush(fp);
		fclose(fp);
	}
#endif

	printf ("====> Exported: %s\n", blkdev);
	printf ("=======> You can flash the device now\n");

	netlink_socket();

	/*
	 * We should never been here unless NETLINK failed
	 */
	while (1) { sleep(10); }
}

